package net.kronos.mockup.handler;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.kronos.mockup.annotation.MockPersonName;
import net.kronos.orm.rowset.ExceptionHandler;

import static net.kronos.mockup.annotation.MockPersonName.Pattern;
import static net.kronos.mockup.annotation.MockPersonName.Nationality;

public class PersonNameHandler implements IMockHandlerCommand{
	public static int KANA_COL = 0;
	public static int NAME_COL = 1;
	public static int DIV_COL = 2;
	public static int DIV_FIRST_NAME = 1;
	public static int DIV_LAST_NAME = 2;
	private static final int NAMELIST_SIZE = 5000;
	private int nameListCreateTimes = 0;
	
	protected int maxListSize = 0;
	
	protected Set<Name> firstNames = new HashSet<Name>();
	protected Set<Name> lastNames = new HashSet<Name>();
	protected List<FullName> namesList;
	protected Map<String, String> kanaRomanTable = PersonNameHandler.getHiraganaToRoman();
	protected Map<String, String> multiKanaRomanTable = PersonNameHandler.getMultiHiragana();
	
	public PersonNameHandler(){
		readNameCSVFile();
		this.namesList = getPersonNameList(NAMELIST_SIZE);
	}
	
	
	@SuppressWarnings("unchecked")
	@Override
	public Object execute(Annotation ann, Field targetField,Object instance,int count) {
		MockPersonName annotation = (MockPersonName)ann;

		// 名前リストの作り直しを考慮したインデックスの作成
		int index = count - (nameListCreateTimes * NAMELIST_SIZE);
		
		// インデックスが名前リストの末尾まできたら、名前リストを作り直してインデックスも0に戻す。
		if(namesList.size() <= index ){
			readNameCSVFile();
			this.namesList = getPersonNameList(NAMELIST_SIZE);
			nameListCreateTimes ++;
			// インデックスを0に戻す。
			index = count - (nameListCreateTimes * NAMELIST_SIZE);
		}
		FullName name = namesList.get(index);
		return name.getNameString(annotation);
	}

	/**
	 * 人物名リストCSVを読み取る。
	 * @return
	 */
	protected void readNameCSVFile(){
		try {
			this.firstNames = new HashSet<Name>();
			this.lastNames = new HashSet<Name>();
			
			InputStream in = this.getClass().getResourceAsStream("/net/kronos/mockup/HumanNameListJp.csv");
			BufferedReader br = new BufferedReader(new InputStreamReader(in));
			
			String line;
			while((line = br.readLine()) != null){
				String[] cols = line.split(",");
				int div = Integer.parseInt(cols[DIV_COL]);
				Name n = new Name(cols[NAME_COL],cols[KANA_COL],div);
				if(div == DIV_FIRST_NAME){
					this.firstNames.add(n);
				}else{
					this.lastNames.add(n);
				}
			}
			br.close();
			in.close();

			return;
		} catch (FileNotFoundException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		} catch (IOException e) {
			throw ExceptionHandler.wrapNoCheckException(e);
		}
	}
	
	/**
	 * 氏名を組み合わせた名前のリストを取得する。
	 * @return 氏名を組み合わせた名前のリスト
	 */
	protected List<FullName> getPersonNameList(int size){
		List<FullName> resultValue = new ArrayList<FullName>();
		Iterator<Name> firstNameIt = firstNames.iterator();
		Iterator<Name> lastNameIt = lastNames.iterator();
		int count = 0;
		while(firstNameIt.hasNext()){
			Name fn = firstNameIt.next();
			Name ln;
			if(lastNameIt.hasNext()){
				ln = lastNameIt.next();
			}else{
				lastNameIt = lastNames.iterator();
				continue;
			}
			resultValue.add(new FullName(fn,ln));
			count++;
			if(count >= size){
				break;
			}
		}
		this.maxListSize = resultValue.size();
		return resultValue;
	}
	
	/**
	 * 人物名の氏名とふり仮名を持つクラス
	 * @author dai.yamamoto
	 */
	public class FullName{
		public FullName(Name firstName, Name lastName) {
			this.firstName = firstName;
			this.lastName = lastName;
		}
		Name firstName;
		Name lastName;
		public String getNameString(MockPersonName annotation){
			Pattern pattern = annotation.pattern();
			Nationality na = annotation.nationality();
			
			if(pattern ==  Pattern.SpacedFullName){
				return firstName.getValue(na) + " " + lastName.getValue(na);
			}else if(pattern == Pattern.FirstName){
				return firstName.getValue(na);
			}else if(pattern == Pattern.LastName){
				return lastName.getValue(na);
			}else if(pattern == Pattern.MailAddress){
				String mailSuffix = annotation.mailAddressAfterAtmark();
				return 
					lastName.getValue(Nationality.JP_Roman) + "_" 
					+ firstName.getValue(Nationality.JP_Roman) 
					+ mailSuffix;
			}else{
				return firstName.getValue(na) + lastName.getValue(na);
			}
		}
	}
	
	/**
	 * 人物名リストCSVを読み込むPojo
	 * @author dai.yamamoto
	 */
	public class Name {
		public Name(String name, String kana, int div) {
			this.name = name;
			this.kana = kana;
			this.div = div;
		}
		public String name;
		public String kana;
		public int div;
		
		/**
		 * ローマ字を返す。
		 * @return
		 */
		private String getRoman(){
			 Set<String> keys = multiKanaRomanTable.keySet();
			 String ret = this.kana;
			 for(String k : keys){
				 ret = ret.replaceAll(k, multiKanaRomanTable.get(k));
			 }
			 keys = kanaRomanTable.keySet();
			 for(String k : keys){
				 ret = ret.replaceAll(k, kanaRomanTable.get(k));
			 }
			 
			 // 吃音「っ」は子音を重ねる。
			 int inx = ret.indexOf("っ");
			 if(inx > 0){
				 if(inx + 1 < ret.length()){
					char rep = ret.charAt(inx + 1);
					ret = ret.replace('っ', rep);
				 }else{
					ret = ret.replace("っ","");
				 }
			 }
			 return ret;
		}
		
		/**
		 * アノテーションのパラメータにしたがって名前を返す。
		 * @param na
		 * @return
		 */
		public String  getValue(Nationality na){
			if(na == Nationality.JP_Kana){
				return kana;
			}else if (na==Nationality.JP_Roman){
				return this.getRoman();
			}else{
				return name;
			}
		}
	}
	 public static Map<String,String> getMultiHiragana(){
			Map<String, String> hiraganaToKunrei = new Hashtable<String, String>();
			hiraganaToKunrei.put("きゃ","kya");
			hiraganaToKunrei.put("きゅ","kyu");
			hiraganaToKunrei.put("きょ","kyo");
			hiraganaToKunrei.put("しゃ","sha");
			hiraganaToKunrei.put("しゅ","shu");
			hiraganaToKunrei.put("しょ","sho");
			hiraganaToKunrei.put("ちゃ","cha");
			hiraganaToKunrei.put("ちゅ","chu");
			hiraganaToKunrei.put("ちょ","cho");
			hiraganaToKunrei.put("にゃ","nya");
			hiraganaToKunrei.put("にゅ","nyu");
			hiraganaToKunrei.put("にょ","nyo");
			hiraganaToKunrei.put("ひゃ","hya");
			hiraganaToKunrei.put("ひゅ","hyu");
			hiraganaToKunrei.put("ひょ","hyo");
			hiraganaToKunrei.put("みゃ","mya");
			hiraganaToKunrei.put("みゅ","myu");
			hiraganaToKunrei.put("みょ","myo");
			hiraganaToKunrei.put("りゃ","rya");
			hiraganaToKunrei.put("りゅ","ryu");
			hiraganaToKunrei.put("りょ","ryo");
			hiraganaToKunrei.put("ぎゃ","gya");
			hiraganaToKunrei.put("ぎゅ","gyu");
			hiraganaToKunrei.put("ぎょ","gyo");
			hiraganaToKunrei.put("じゃ","ja");
			hiraganaToKunrei.put("じゅ","ju");
			hiraganaToKunrei.put("じょ","jo");
			hiraganaToKunrei.put("びゃ","bya");
			hiraganaToKunrei.put("びゅ","byu");
			hiraganaToKunrei.put("びょ","byo");
			hiraganaToKunrei.put("ぴゃ","pya");
			hiraganaToKunrei.put("ぴゅ","pyu");
			hiraganaToKunrei.put("ぴょ","pyo");
			return hiraganaToKunrei;
	 }
    private static Map<String, String> getHiraganaToRoman() {
		Map<String, String> hiraganaToKunrei = new Hashtable<String, String>();

		
		hiraganaToKunrei.put("あ","a");
		hiraganaToKunrei.put("い","i");
		hiraganaToKunrei.put("う","u");
		hiraganaToKunrei.put("え","e");
		hiraganaToKunrei.put("お","o");
		hiraganaToKunrei.put("か","ka");
		hiraganaToKunrei.put("き","ki");
		hiraganaToKunrei.put("く","ku");
		hiraganaToKunrei.put("け","ke");
		hiraganaToKunrei.put("こ","ko");
		hiraganaToKunrei.put("さ","sa");
		hiraganaToKunrei.put("し","shi");
		hiraganaToKunrei.put("す","su");
		hiraganaToKunrei.put("せ","se");
		hiraganaToKunrei.put("そ","so");
		hiraganaToKunrei.put("た","ta");
		hiraganaToKunrei.put("ち","chi");
		hiraganaToKunrei.put("つ","tsu");
		hiraganaToKunrei.put("て","te");
		hiraganaToKunrei.put("と","to");
		hiraganaToKunrei.put("な","na");
		hiraganaToKunrei.put("に","ni");
		hiraganaToKunrei.put("ぬ","nu");
		hiraganaToKunrei.put("ね","ne");
		hiraganaToKunrei.put("の","no");
		hiraganaToKunrei.put("は","ha");
		hiraganaToKunrei.put("ひ","hi");
		hiraganaToKunrei.put("ふ","fu");
		hiraganaToKunrei.put("へ","he");
		hiraganaToKunrei.put("ほ","ho");
		hiraganaToKunrei.put("ま","ma");
		hiraganaToKunrei.put("み","mi");
		hiraganaToKunrei.put("む","mu");
		hiraganaToKunrei.put("め","me");
		hiraganaToKunrei.put("も","mo");
		hiraganaToKunrei.put("や","ya");
		hiraganaToKunrei.put("ゆ","yu");
		hiraganaToKunrei.put("ゐ","e");
		hiraganaToKunrei.put("よ","yo");
		hiraganaToKunrei.put("ら","ra");
		hiraganaToKunrei.put("り","ri");
		hiraganaToKunrei.put("る","ru");
		hiraganaToKunrei.put("れ","re");
		hiraganaToKunrei.put("ろ","ro");
		hiraganaToKunrei.put("わ","wa");
		hiraganaToKunrei.put("ゐ","e");
		hiraganaToKunrei.put("ゑ","we");
		hiraganaToKunrei.put("を","wo");
		hiraganaToKunrei.put("ん","n");
		hiraganaToKunrei.put("が","ga");
		hiraganaToKunrei.put("ぎ","gi");
		hiraganaToKunrei.put("ぐ","gu");
		hiraganaToKunrei.put("げ","ge");
		hiraganaToKunrei.put("ご","go");
		hiraganaToKunrei.put("ざ","za");
		hiraganaToKunrei.put("じ","ji");
		hiraganaToKunrei.put("ず","zu");
		hiraganaToKunrei.put("ぜ","ze");
		hiraganaToKunrei.put("ぞ","zo");
		hiraganaToKunrei.put("だ","da");
		hiraganaToKunrei.put("ぢ","di");
		hiraganaToKunrei.put("づ","du");
		hiraganaToKunrei.put("で","de");
		hiraganaToKunrei.put("ど","do");
		hiraganaToKunrei.put("ば","ba");
		hiraganaToKunrei.put("び","bi");
		hiraganaToKunrei.put("ぶ","bu");
		hiraganaToKunrei.put("べ","be");
		hiraganaToKunrei.put("ぼ","bo");
		hiraganaToKunrei.put("ぱ","pa");
		hiraganaToKunrei.put("ぴ","pi");
		hiraganaToKunrei.put("ぷ","pu");
		hiraganaToKunrei.put("ぺ","pe");
		hiraganaToKunrei.put("ぽ","po");
		
        return hiraganaToKunrei;
    }
}
